@@ -3,8 +3,8 @@ use magnus::prelude::*;
 use magnus::rb_sys::{AsRawValue, FromRawValue};
 use magnus::typed_data::Obj;
 use magnus::value::Fixnum;
-use magnus::{block::Proc, eval, scan_args, Error, Integer, RArray, RClass, RHash, RString, Value};
-use nix::libc::c_long;
+use magnus::{eval, scan_args, Error, Integer, RArray, RClass, RHash, RString, Value};
+use nix::libc::{c_char, c_long, c_ulong};
 use rb_sys::rb_str_new_static;
 use std::fs::File;
 use std::io::{prelude::*, SeekFrom};
@@ -25,6 +25,18 @@ use inner::InnerMmap;
 
 mod inner;
 
+#[cfg(ruby_gte_3_4)]
+/// The Ruby `STR_SHARED` flag, aka `FL_USER0`.
+/// This was changed from `FL_USER2` in https://github.com/ruby/ruby/commit/6deeec5d459ecff5ec4628523b14ac7379fd942e.
+const STR_SHARED: c_ulong = 1 << (12);
+
+#[cfg(ruby_lte_3_3)]
+/// The Ruby `STR_SHARED` flag, aka `FL_USER2`.
+const STR_SHARED: c_ulong = 1 << (14);
+
+/// The Ruby `STR_NOEMBED` flag, aka `FL_USER1`.
+const STR_NOEMBED: c_ulong = 1 << (13);
+
 /// A Rust struct wrapped in a Ruby object, providing access to a memory-mapped
 /// file used to store, update, and read out Prometheus metrics.
 ///
@@ -151,13 +163,28 @@ impl MmapedFile {
     ///
     /// return a substring of <em>lenght</em> characters from <em>start</em>
     pub fn slice(rb_self: Obj<Self>, args: &[Value]) -> magnus::error::Result<RString> {
-        // The C implemenation would trigger a GC cycle via `rb_gc_force_recycle`
+        // The C implementation would trigger a GC cycle via `rb_gc_force_recycle`
         // if the `MM_PROTECT` flag is set, but in practice this is never used.
         // We omit this logic, particularly because `rb_gc_force_recycle` is a
         // no-op as of Ruby 3.1.
+        let rs_self = &*rb_self;
 
-        let str = (*rb_self).str(rb_self)?;
-        let res = str.funcall("[]", args);
+        let str = rs_self.str(rb_self)?;
+        rs_self._slice(rb_self, str, args)
+    }
+
+    fn _slice(
+        &self,
+        rb_self: Obj<Self>,
+        str: RString,
+        args: &[Value],
+    ) -> magnus::error::Result<RString> {
+        let substr: RString = str.funcall("[]", args)?;
+
+        // Track shared child strings which use the same backing storage.
+        if Self::rb_string_is_shared(substr) {
+            (*rb_self).track_rstring(rb_self, substr)?;
+        }
 
         // The C implementation does this, perhaps to validate that the len we
         // provided is actually being used.
@@ -166,7 +193,7 @@ impl MmapedFile {
             Ok(())
         })?;
 
-        res
+        Ok(substr)
     }
 
     /// Document-method: msync
@@ -205,9 +232,6 @@ impl MmapedFile {
         let rs_self = &*rb_self;
 
         rs_self.inner_mut(|inner| {
-            // truncate file to actual used size
-            inner.truncate_file()?;
-
             // We are about to release the backing mmap for Ruby's String
             // objects. If Ruby attempts to read from them the program will
             // segfault. We update the length of all Strings to zero so Ruby
@@ -223,7 +247,8 @@ impl MmapedFile {
         })?;
 
         // Update each String object to be zero-length.
-        rs_self.update_weak_map(rb_self)?;
+        let cap = util::cast_chk::<_, c_long>(rs_self.capacity(), "capacity")?;
+        rs_self.update_weak_map(rb_self, rs_self.as_mut_ptr(), cap)?;
 
         // Remove the `InnerMmap` from the `RwLock`. This will drop
         // end of this function, unmapping and closing the file.
@@ -341,16 +366,19 @@ impl MmapedFile {
             Ok(unsafe { rb_str_new_static(ptr as _, len as _) })
         })?;
 
+        // SAFETY: We know that rb_str_new_static returns a VALUE.
         let val = unsafe { Value::from_raw(val_id) };
 
         // UNWRAP: We created this value as a string above.
         let str = RString::from_value(val).unwrap();
 
-        let tracker: Value = rb_self.ivar_get("@weak_obj_tracker")?;
+        // Freeze the root string so it can't be mutated out from under any
+        // substrings created. This object is never exposed to callers.
+        str.freeze();
 
-        // Use the string's Id as the key in the `WeakMap`.
-        let key = str.as_raw();
-        let _: Value = tracker.funcall("[]=", (key, str))?;
+        // Track the RString in our `WeakMap` so we can update its address if
+        // we re-mmap the backing file.
+        (*rb_self).track_rstring(rb_self, str)?;
 
         Ok(str)
     }
@@ -358,23 +386,24 @@ impl MmapedFile {
     /// If we reallocate, any live Ruby strings provided by the `str()` method
     /// will be invalidated. We need to iterate over them using and update their
     /// heap pointers to the newly allocated memory region.
-    fn update_weak_map(&self, rb_self: Obj<Self>) -> magnus::error::Result<()> {
+    fn update_weak_map(
+        &self,
+        rb_self: Obj<Self>,
+        old_ptr: *const c_char,
+        old_cap: c_long,
+    ) -> magnus::error::Result<()> {
         let tracker: Value = rb_self.ivar_get("@weak_obj_tracker")?;
 
-        let (new_ptr, new_len) = (*rb_self).inner(|inner| {
-            // Pointers are not `Send`, convert it to usize to allow capture in closure.
-            let new_ptr = inner.as_ptr() as usize;
-            let new_len = util::cast_chk::<_, c_long>(inner.len(), "mmap len")?;
+        let new_len = self.inner(|inner| util::cast_chk::<_, c_long>(inner.len(), "mmap len"))?;
 
-            Ok((new_ptr, new_len))
-        })?;
+        // Convert WeakMap to an Array of values to to avoid garbage collection issues
+        let values: Value = tracker.funcall("values", ())?;
 
-        // Allocate a block with a closure that updates the string details.
-        let block = Proc::from_fn(move |args, _block| {
-            let val = args
-                .get(0)
-                .ok_or_else(|| err!(arg_error(), "no argument received from `each_value`"))?;
-            let str = RString::from_value(*val)
+        // Iterate over the values of the `WeakMap`.
+        for val in values.enumeratorize("each", ()) {
+            let rb_string = val?;
+
+            let str = RString::from_value(rb_string)
                 .ok_or_else(|| err!(arg_error(), "weakmap value was not a string"))?;
 
             // SAFETY: We're messing with Ruby's internals here, YOLO.
@@ -383,18 +412,42 @@ impl MmapedFile {
                 // which provides access to its internals.
                 let mut raw_str = Self::rb_string_internal(str);
 
+                // Shared string have their own `ptr` and `len` values, but `aux`
+                // is the id of the parent string so the GC can track this
+                // dependency. The `ptr` will always be an offset from the base
+                // address of the mmap, and `len` will be the length of the mmap
+                // less the offset from the base.
+                if Self::rb_string_is_shared(str) && new_len > 0 {
+                    // Calculate how far into the original mmap the shared string
+                    // started and update to the equivalent address in the new
+                    // one.
+                    let substr_ptr = raw_str.as_ref().as_.heap.ptr;
+                    let offset = substr_ptr.offset_from(old_ptr);
+
+                    raw_str.as_mut().as_.heap.ptr = self.as_mut_ptr().offset(offset);
+
+                    let current_len = str.len() as c_long;
+                    let new_shared_len = old_cap + current_len;
+
+                    self.update_rstring_len(raw_str, new_shared_len);
+                    continue;
+                }
+
                 // Update the string to point to the new mmapped file.
                 // We're matching the behavior of Ruby's `str_new_static` function.
                 // See https://github.com/ruby/ruby/blob/e51014f9c05aa65cbf203442d37fef7c12390015/string.c#L1030-L1053
-                raw_str.as_mut().as_.heap.ptr = new_ptr as _;
-                raw_str.as_mut().as_.heap.len = new_len;
-                raw_str.as_mut().as_.heap.aux.capa = new_len;
+                //
+                // We deliberately do _NOT_ increment the `capa` field of the
+                // string to match the new `len`. We were initially doing this,
+                // but consistently triggered GCs in the middle of updating the
+                // string pointers, causing a segfault.
+                //
+                // See https://gitlab.com/gitlab-org/ruby/gems/prometheus-client-mmap/-/issues/45
+                raw_str.as_mut().as_.heap.ptr = self.as_mut_ptr();
+                self.update_rstring_len(raw_str, new_len);
             }
-            Ok(())
-        });
+        }
 
-        // Execute the block.
-        let _: Value = tracker.funcall_with_block("each_value", (), block)?;
         Ok(())
     }
 
@@ -429,6 +482,9 @@ impl MmapedFile {
         }
 
         if new_cap != self.capacity() {
+            let old_ptr = self.as_mut_ptr();
+            let old_cap = util::cast_chk::<_, c_long>(self.capacity(), "capacity")?;
+
             // Drop the old mmap.
             let (mut file, path) = self.take_inner()?.munmap();
 
@@ -439,7 +495,7 @@ impl MmapedFile {
 
             self.insert_inner(new_inner)?;
 
-            return self.update_weak_map(rb_self);
+            return self.update_weak_map(rb_self, old_ptr, old_cap);
         }
 
         Ok(())
@@ -459,14 +515,14 @@ impl MmapedFile {
         match file.seek(SeekFrom::Start(len - 1)) {
             Ok(_) => {}
             Err(_) => {
-                return Err(MmapError::WithErrno(format!("Can't lseek {}", len - 1)));
+                return Err(MmapError::with_errno(format!("Can't lseek {}", len - 1)));
             }
         }
 
         match file.write(&[0x0]) {
             Ok(1) => {}
             _ => {
-                return Err(MmapError::WithErrno(format!(
+                return Err(MmapError::with_errno(format!(
                     "Can't extend {}",
                     path.display()
                 )));
@@ -476,6 +532,15 @@ impl MmapedFile {
         Ok(())
     }
 
+    fn track_rstring(&self, rb_self: Obj<Self>, str: RString) -> magnus::error::Result<()> {
+        let tracker: Value = rb_self.ivar_get("@weak_obj_tracker")?;
+
+        // Use the string's Id as the key in the `WeakMap`.
+        let key = str.as_raw();
+        let _: Value = tracker.funcall("[]=", (key, str))?;
+        Ok(())
+    }
+
     /// The total capacity of the underlying mmap.
     #[inline]
     fn capacity(&self) -> usize {
@@ -489,6 +554,13 @@ impl MmapedFile {
             .map_err(|e| e.into())
     }
 
+    fn as_mut_ptr(&self) -> *mut c_char {
+        // UNWRAP: This is actually infallible, but we need to
+        // wrap it in a `Result` for use with `inner()`.
+        self.inner(|inner| Ok(inner.as_mut_ptr() as *mut c_char))
+            .unwrap()
+    }
+
     /// Takes a closure with immutable access to InnerMmap. Will fail if the inner
     /// object has a mutable borrow or has been dropped.
     fn inner<F, T>(&self, func: F) -> Result<T>
@@ -544,11 +616,39 @@ impl MmapedFile {
         Ok(())
     }
 
+    /// Check if an RString is shared. Shared string use the same underlying
+    /// storage as their parent, taking an offset from the start. By default
+    /// they must run to the end of the parent string.
+    fn rb_string_is_shared(rb_str: RString) -> bool {
+        // SAFETY: We only hold a reference to the raw object for the duration
+        // of this function, and no Ruby code is called.
+        let flags = unsafe {
+            let raw_str = Self::rb_string_internal(rb_str);
+            raw_str.as_ref().basic.flags
+        };
+        let shared_flags = STR_SHARED | STR_NOEMBED;
+
+        flags & shared_flags == shared_flags
+    }
+
     /// Convert `magnus::RString` into the raw binding used by `rb_sys::RString`.
     /// We need this to manually change the pointer and length values for strings
     /// when moving the mmap to a new file.
+    ///
+    /// SAFETY: Calling Ruby code while the returned object is held may result
+    /// in it being mutated or dropped.
     unsafe fn rb_string_internal(rb_str: RString) -> NonNull<rb_sys::RString> {
-        mem::transmute::<RString, NonNull<rb_sys::RString>>(rb_str)
+        NonNull::new_unchecked(rb_str.as_raw() as *mut rb_sys::RString)
+    }
+
+    #[cfg(ruby_lte_3_2)]
+    unsafe fn update_rstring_len(&self, mut raw_str: NonNull<rb_sys::RString>, new_len: c_long) {
+        raw_str.as_mut().as_.heap.len = new_len;
+    }
+
+    #[cfg(ruby_gte_3_3)]
+    unsafe fn update_rstring_len(&self, mut raw_str: NonNull<rb_sys::RString>, new_len: c_long) {
+        raw_str.as_mut().len = new_len;
     }
 }
 
@@ -666,18 +766,100 @@ mod test {
         let ruby = magnus::Ruby::get().unwrap();
         crate::init(&ruby).unwrap();
 
+        fn assert_internals(
+            obj: Obj<MmapedFile>,
+            parent_id: c_ulong,
+            child_id: c_ulong,
+            unshared_id: c_ulong,
+        ) {
+            let rs_self = &*obj;
+            let tracker: Value = obj.ivar_get("@weak_obj_tracker").unwrap();
+
+            let mmap_ptr = rs_self.as_mut_ptr();
+            let mmap_len = rs_self.capacity();
+
+            let mut parent_checked = false;
+            let mut child_checked = false;
+
+            for val in tracker.enumeratorize("each_value", ()) {
+                let rb_string = val.unwrap();
+                let str = RString::from_value(rb_string).unwrap();
+
+                unsafe {
+                    let raw_str = MmapedFile::rb_string_internal(str);
+                    if str.as_raw() == child_id {
+                        assert_eq!(parent_id, raw_str.as_ref().as_.heap.aux.shared);
+
+                        let child_offset = mmap_len as isize - str.len() as isize;
+                        assert_eq!(mmap_ptr.offset(child_offset), raw_str.as_ref().as_.heap.ptr);
+
+                        child_checked = true;
+                    } else if str.as_raw() == parent_id {
+                        assert_eq!(parent_id, str.as_raw());
+
+                        assert_eq!(mmap_ptr, raw_str.as_ref().as_.heap.ptr);
+                        assert_eq!(mmap_len as c_long, str.len() as c_long);
+                        assert!(raw_str.as_ref().basic.flags & (STR_SHARED | STR_NOEMBED) > 0);
+                        assert!(str.is_frozen());
+
+                        parent_checked = true;
+                    } else if str.as_raw() == unshared_id {
+                        panic!("tracking unshared string");
+                    } else {
+                        panic!("unknown string");
+                    }
+                }
+            }
+            assert!(parent_checked && child_checked);
+        }
+
         let obj = create_obj();
         let _ = populate_entries(&obj);
 
         let rs_self = &*obj;
 
-        let value_range = Range::new(HEADER_SIZE, 24, false).unwrap().as_value();
-        let value_slice = MmapedFile::slice(obj, &[value_range]).unwrap();
+        // Create a string containing the full mmap.
+        let parent_str = rs_self.str(obj).unwrap();
+        let parent_id = parent_str.as_raw();
 
-        rs_self.expand_to_fit(obj, rs_self.capacity() * 2).unwrap();
+        // Ruby's shared strings are only created when they go to the end of
+        // original string.
+        let len = rs_self.inner(|inner| Ok(inner.len())).unwrap();
+        let shareable_range = Range::new(1, len - 1, false).unwrap().as_value();
+
+        // This string should re-use the parent's buffer with an offset and have
+        // the parent's id in `as.heap.aux.shared`
+        let child_str = rs_self._slice(obj, parent_str, &[shareable_range]).unwrap();
+        let child_id = child_str.as_raw();
+
+        // A range that does not reach the end of the parent will not be shared.
+        assert!(len > 4);
+        let unshareable_range = Range::new(0, 4, false).unwrap().as_value();
+
+        // This string should NOT be tracked, it should own its own buffer.
+        let unshared_str = rs_self
+            ._slice(obj, parent_str, &[unshareable_range])
+            .unwrap();
+        let unshared_id = unshared_str.as_raw();
+        assert!(!MmapedFile::rb_string_is_shared(unshared_str));
+
+        assert_internals(obj, parent_id, child_id, unshared_id);
+
+        let orig_ptr = rs_self.as_mut_ptr();
+        // Expand a bunch to ensure we remap
+        for _ in 0..16 {
+            rs_self.expand_to_fit(obj, rs_self.capacity() * 2).unwrap();
+        }
+        let new_ptr = rs_self.as_mut_ptr();
+        assert!(orig_ptr != new_ptr);
 
         // If we haven't updated the pointer to the newly remapped file this will segfault.
-        let _: Value = eval!("puts str", str = value_slice).unwrap();
+        let _: Value = eval!("puts parent", parent = parent_str).unwrap();
+        let _: Value = eval!("puts child", child = child_str).unwrap();
+        let _: Value = eval!("puts unshared", unshared = unshared_str).unwrap();
+
+        // Confirm that tracked strings are still valid.
+        assert_internals(obj, parent_id, child_id, unshared_id);
     }
 
     #[test]
