@@ -204,6 +204,12 @@ impl InnerMmap {
         self.map.as_ptr()
     }
 
+    /// Returns a mutable raw pointer to the mmap.
+    /// For use in updating RString internals which requires a mutable pointer.
+    pub fn as_mut_ptr(&self) -> *mut u8 {
+        self.map.as_ptr().cast_mut()
+    }
+
     /// Perform an msync(2) on the mmap, flushing all changes written
     /// to disk. The sync may optionally be performed asynchronously.
     pub fn flush(&mut self, f_async: bool) -> Result<()> {
@@ -218,16 +224,6 @@ impl InnerMmap {
         }
     }
 
-    /// Truncate the mmapped file to the end of the metrics data.
-    pub fn truncate_file(&mut self) -> Result<()> {
-        // CAST: no-op on 64-bit, widening on 32-bit.
-        let trunc_len = self.len as u64;
-
-        self.file
-            .set_len(trunc_len)
-            .map_err(|e| MmapError::legacy(format!("truncate: {e}"), RubyError::Type))
-    }
-
     /// Load the `used` header containing the size of the metrics data written.
     pub fn load_used(&self) -> Result<u32> {
         match read_u32(self.map.as_ref(), 0) {
