@@ -1,5 +1,6 @@
 require 'prometheus/client/registry'
 require 'prometheus/client/mmaped_value'
+require 'prometheus/client/page_size'
 require 'logger'
 require 'tmpdir'
 
@@ -10,7 +11,7 @@ module Prometheus
 
       def initialize
         @value_class = ::Prometheus::Client::MmapedValue
-        @initial_mmap_file_size = 4 * 1024
+        @initial_mmap_file_size = ::Prometheus::Client::PageSize.page_size(fallback_page_size: 4096)
         @logger = Logger.new($stdout)
         @pid_provider = Process.method(:pid)
         @multiprocess_files_dir = ENV.fetch('prometheus_multiproc_dir') do
