@@ -26,49 +26,16 @@ module Prometheus
             Helper::MetricsRepresentation.to_text(metrics)
           end
 
-          def marshal_multiprocess(path = Prometheus::Client.configuration.multiprocess_files_dir, use_rust: false)
+          def marshal_multiprocess(path = Prometheus::Client.configuration.multiprocess_files_dir)
             file_list = Dir.glob(File.join(path, '*.db')).sort
               .map {|f| Helper::PlainFile.new(f) }
               .map {|f| [f.filepath, f.multiprocess_mode.to_sym, f.type.to_sym, f.pid] }
 
-            if use_rust && rust_impl_available?
               FastMmapedFileRs.to_metrics(file_list.to_a)
-            else
-              FastMmapedFile.to_metrics(file_list.to_a)
-            end
-          end
-
-          def rust_impl_available?
-            return @rust_available unless @rust_available.nil?
-
-            check_for_rust
           end
 
           private
 
-          def load_rust_extension
-            begin
-              ruby_version = /(\d+\.\d+)/.match(RUBY_VERSION)
-              require_relative "../../../#{ruby_version}/fast_mmaped_file_rs"
-            rescue LoadError
-              require 'fast_mmaped_file_rs'
-            end
-          end
-
-          def check_for_rust
-            # This will be evaluated on each invocation even with `||=` if
-            # `@rust_available` if false. Running a `require` statement is slow,
-            # so the `rust_impl_available?` method memoizes the result, external
-            # callers can only trigger this method a single time.
-            @rust_available = begin
-              load_rust_extension
-              true
-            rescue LoadError
-              Prometheus::Client.logger.info('FastMmapedFileRs unavailable')
-              false
-            end
-          end
-
           def load_metrics(path)
             metrics = {}
             Dir.glob(File.join(path, '*.db')).sort.each do |f|
