@@ -31,6 +31,10 @@ module Prometheus
       def increment(labels, value)
         @values[label_set_for(labels)].increment(value)
       end
+
+      def decrement(labels, value)
+        @values[label_set_for(labels)].decrement(value)
+      end
     end
   end
 end
