@@ -4,15 +4,15 @@ require 'prometheus/client/helper/file_locker'
 # load precompiled extension if available
 begin
   ruby_version = /(\d+\.\d+)/.match(RUBY_VERSION)
-  require_relative "../../../#{ruby_version}/fast_mmaped_file"
+  require_relative "../../../#{ruby_version}/fast_mmaped_file_rs"
 rescue LoadError
-  require 'fast_mmaped_file'
+  require 'fast_mmaped_file_rs'
 end
 
 module Prometheus
   module Client
     module Helper
-      class MmapedFile < FastMmapedFile
+      class MmapedFile < FastMmapedFileRs
         include EntryParser
 
         attr_reader :filepath, :size
