@@ -11,6 +11,9 @@ describe Prometheus::Client::MmapedValue, :temp_metrics_dir do
     let(:counter) { described_class.new(:counter, :counter, 'counter', {}) }
 
     before do
+      # reset all current metrics
+      Prometheus::Client::MmapedValue.class_variable_set(:@@files, {})
+
       counter.increment(1)
     end
 
